from sqlalchemy import *

from dataProviderBase import *
from T0Mon.utils import *
from T0Mon.analysis import *
from T0Mon.consts import *
from T0Mon.sqlManager import *

from T0Mon.globals import logger
import T0Mon.config

### Module for classes providing data about T0 runs ###

class RunConfigKeyProvider(TableDataProvider):
    col1 = [Col("Dataset"), Cols([])]
    col2 = [Col("Repack"), Cols([Col("Proc Version"), Col("HLTDebug")])]
    col3 = [Col("Reco"), Cols([Col("Reco"), Col("CMSsw Version"), Col("Global tag"), Col("Config URL", decorator=buildWebLink), Col("Proc Version")])]
    col4 = [Col("Alcaskim"), Cols([Col("Alcaskim"), Col("CMSsw Version"), Col("Config URL", decorator=buildWebLink), Col("Proc Version")])]
    columns = []
    columns.append(col1)
    columns.append(col2)
    columns.append(col3)
    columns.append(col4)

    cols = DoubleCols(columns)

    def __init__(self, runid):
        self.runid = runid
        self.title = "Run " + str(self.runid)
        TableDataProvider.__init__(self)
        TableDataProvider._load(self)

    def _buildQuery(self):
        repackConfig = Tbl(TABLES.REPACK_CONFIG, True)
        recoConfig = Tbl(TABLES.RECO_CONFIG, True)
	alcaConfig = Tbl(TABLES.ALCA_CONFIG, True)
        primaryDataset = Tbl(TABLES.PRIMARY_DATASET, True)
        cmsswVersion = Tbl(TABLES.CMSSW_VERSION, True)
        
        if self.asc:
            order = "ASC"
        else:
            order = "DESC"


        self.query ="""
        SELECT sub.ds_dataset, sub.rp_proc_version, case sub.rp_hltdebug when 0 then 'Disabled' when 1 then 'Enabled' end rp_hltdebug, case sub.ro_reco when 0 then 'Disabled' when 1 then 'Enabled' end ro_reco, sub.ro_cmssw_version, sub.ro_global_tag, sub.ro_config_url, sub.ro_proc_version, case sub.al_alcaskim when 0 then 'Disabled' when 1 then 'Enabled' end al_alcaskim, sub.al_cmssw_version, sub.al_config_url, sub.al_proc_version 
        FROM(
               SELECT
               ds.name ds_dataset,
               rp.proc_version rp_proc_version,
               rp.hltdebug rp_hltdebug,
               ro.do_reco ro_reco,
               sw1.name ro_cmssw_version,
               ro.global_tag ro_global_tag,
               ro.config_url ro_config_url,
               ro.proc_version ro_proc_version,
	       al.do_alca al_alcaskim,
	       sw2.name al_cmssw_version,
	       al.config_url al_config_url,
	       al.proc_version al_proc_version
               FROM """ + repackConfig + """ rp, """ + recoConfig + """ ro, """ + alcaConfig  + """ al, """  + cmsswVersion  + """ sw1, """ + cmsswVersion  + """ sw2, """ + primaryDataset  + """ ds
               WHERE rp.run_id = """ + str(self.runid) + """
               AND rp.run_id = ro.run_id AND ro.run_id = al.run_id  AND rp.primary_dataset_id = ro.primary_dataset_id AND ro.primary_dataset_id = al.primary_dataset_id AND ro.cmssw_version_id = sw1.id AND al.cmssw_version_id = sw2.id  AND rp.primary_dataset_id = ds.id
               ORDER BY """ + str(int(self.orderby)+1) + " " + order + """
        ) sub
        """
        

class RunsDataProvider(PagedTableDataProvider):
    RUN_ID_TITLE = "Run ID"
    STREAMERS_TITLE = "Streamers"
    REPACKED_TITLE = "Repacked"
    RECO_TITLE = "Reconstructed"
    REPACK_MERGE_TITLE = "Merged"
    ALCASKIM_TITLE = "Alcaskimmed"

    cols = Cols([
        Col(RUN_ID_TITLE, decorator=formRunLink),
        Col("Start", decorator=timeToClock),
        Col("End", decorator=timeToClock),
        Col("Process"),
        Col("Version"),
        Col("HLTKey"),
        Col("ACQ_era"),
        Col("Status", info=
            """
                <i style="color:blue;font-size:10px">Active</i> - Run is ongoing and streamer data is still arriving for it<br />
                <i style="color:blue;font-size:10px">CloseOutSegmentInjector</i> - The Tier-0 is trying to make consistent luminosity sections to be repacked (currently we need 4*n streamer files per Luminosity Section)<br />
                <i style="color:blue;font-size:10px">CloseOutScheduler</i> - There is streamer data for the run that has not been assigned to a repack job yet<br />
                <i style="color:blue;font-size:10px">CloseOutRepack</i> - There is streamer data in a run that has not been repacked yet (create repack jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutRepackMerge</i> - There is RAW data in the run that needs to be merged (create repack merge jobs and wait from them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutPromptReco</i> - There is RAW data in the run that needs to be reconstructed (create prompt reco jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">Complete</i> - We have finished processing this run (there could have been failures along the way but we've exhausted all automatic retries)
            """
        ),
        Col(STREAMERS_TITLE, info=
            """
            The number of streamer files
            """
        ),
        Col(REPACKED_TITLE, info=
            """
                The status bar and percentage displays how many streamer
                files were repacked<br /><br />The number shows how many repack files were produced
            """
        ),
        Col(REPACK_MERGE_TITLE, info=
            """
                The upper status bar and the first percentage displays how many repacked
                files were merged<br />The number shows how many merge files were produced
		<br /><br />
		The second status bar and percentage displays how many repacked and merged
		files have already been exported.
            """
        ),
        Col(RECO_TITLE, info=
            """
                The upper status bar and the first percentage displays how many merged
                files, marked for reconstruction, were actually reconstructed<br />
		The number shows how many reco files were produced<br /><br />
		The second bar and percentage displays how many reconstructed files have already been exported
            """
        ),
        Col(ALCASKIM_TITLE, info=
            """
                The upper status bar and the first percentage displays how many files, 
		marked to be alcaskimmed, were actually alcaskimmed<br />
                The number shows how many alcaskim files were produced<br /><br />
                The second bar and percentage displays how many alcaskimmed files have already been exported
            """
        ),
    ])

    def __init__(self, orderby = 0, asc = True, page = 0, pageSize = 10):
        PagedTableDataProvider.__init__(self, orderby, asc, page, pageSize)
        self.title = "Runs"
        PagedTableDataProvider._load(self)

    def _analysis(self):
        for line in self.data:
            runid = line[self.cols.getId(self.RUN_ID_TITLE)]
            numStreamers = line[self.cols.getId(self.STREAMERS_TITLE)]
            numRepacked = line[self.cols.getId(self.REPACKED_TITLE)]
            numMerge = line[self.cols.getId(self.REPACK_MERGE_TITLE)]
            numReco = line[self.cols.getId(self.RECO_TITLE)]
            numAlcaskim = line[self.cols.getId(self.ALCASKIM_TITLE)]
	    numExportedRepackMerge = line[self.cols.getId(self.RECO_TITLE)+1]
	    numExportedReco = line[self.cols.getId(self.RECO_TITLE)+2]
            numExportedAlcaskim = line[self.cols.getId(self.RECO_TITLE)+3]
            
	    # analyse repacked percentage
            repackedP = formatPercentage(numRepacked, getRunRepackedPercentage(runid, numStreamers))
            
	    # analyse merged percentage
	    mergedP = formatDoublePercentage(numMerge, getRunRepackMergePercentage(runid, numRepacked), numExportedRepackMerge, getRunExportedRepackMergePercentage(runid, numRepacked, numExportedRepackMerge), "avail.", "exp.")
            
	    # analyse reco percentage
	    recoP = formatDoublePercentage(numReco, getRunRecoPercentage(runid, numReco), numExportedReco, getRunRecoExportedPercentage(runid, numReco, numExportedReco), "reco.", "exp.")

            # analyse alcaskim percentage
	    alcaskimP = formatDoublePercentage(numAlcaskim, getRunAlcaskimPercentage(runid, numAlcaskim), numExportedAlcaskim, getRunAlcaskimExportedPercentage(runid, numAlcaskim, numExportedAlcaskim), "alca.", "exp.")

            line[self.cols.getId(self.REPACK_MERGE_TITLE)] = mergedP
            line[self.cols.getId(self.REPACKED_TITLE)] = repackedP
            line[self.cols.getId(self.RECO_TITLE)] = recoP
            line[self.cols.getId(self.ALCASKIM_TITLE)] = alcaskimP

	    line = line[:-3]

    def _buildCountQuery(self):
        runTable = Tbl(TABLES.RUN)

        self.countQuery = select(
            [func.count(runTable.c.run_id)]
        )

    def _buildMainQuery(self):
        runTable = Tbl(TABLES.RUN, True)
        runStatusTable = Tbl(TABLES.RUN_STATUS, True)
        streamerTable = Tbl(TABLES.STREAMER, True)
	repackStreamerAssocTable = Tbl(TABLES.REPACK_STREAMER_ASSOC, True)
	wmbsFileDetailsTable = Tbl(TABLES.WMBS_FILE_DETAILS, True)
	wmbsDatasetAssocTable = Tbl(TABLES.WMBS_FILE_DATASET_PATH_ASSOC, True)
	datasetTable = Tbl(TABLES.DATASET_PATH, True)
	dataTierTable = Tbl(TABLES.DATA_TIER, True)
	wmbsFileRunlumiMapTable = Tbl(TABLES.WMBS_FILE_RUNLUMI_MAP, True)
	wmbsSubFilesCompleteTable = Tbl(TABLES.WMBS_SUB_FILES_COMPLETE, True)
	wmbsSubscriptionTable = Tbl(TABLES.WMBS_SUBSCRIPTION, True)
	wmbsSubsTypeTable = Tbl(TABLES.WMBS_SUBS_TYPE, True)
	wmbsFilesetTable = Tbl(TABLES.WMBS_FILESET, True)
	wmbsWorkflowTable = Tbl(TABLES.WMBS_WORKFLOW, True)

        if self.asc: order = "ASC"
        else: order = "DESC"

        # HACK: query time optimisation
        if self.orderby < self.cols.getId(self.STREAMERS_TITLE):
            self.mainQuery = """
            SELECT sub.run_id, sub.start_time, sub.end_time, sub.process, sub.version, sub.hltkey, sub.acq_era, sub.status,

            (SELECT COUNT(*) FROM """ + streamerTable + """ s WHERE s.run_id = sub.run_id) numStreamers,

            (SELECT COUNT(DISTINCT f.id) 
		FROM """ + wmbsFileDetailsTable + """ f INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f.id)
			INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
			INNER JOIN """ + dataTierTable + """ dt ON (dp.data_tier = dt.id)
			INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
		WHERE dt.name = 'RAW' and m.run = sub.run_id) numRepacked,

	    (SELECT COUNT(DISTINCT f.id) 
		FROM """ + wmbsFileDetailsTable + """ f INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f.id)
			INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                        INNER JOIN """ + dataTierTable + """ dt ON (dp.data_tier = dt.id)
                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
			INNER JOIN """ + wmbsSubFilesCompleteTable + """ c ON (f.id = c.fileid)
		WHERE dt.name = 'RAW' and  m.run = sub.run_id and c.subscription = (SELECT id 
											FROM """ + wmbsSubscriptionTable + """
											WHERE fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Mergeable')
											)
		) numMerged, 

            (SELECT COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f 
                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsSubFilesCompleteTable + """ c ON (f.id = c.fileid)
                WHERE m.run = sub.run_id and c.subscription = (SELECT id
                                                                  FROM """ + wmbsSubscriptionTable + """
                                                                  WHERE fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Reconstructable')
                                                                  )
                ) numReconstructed,

            (SELECT COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f
                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsSubFilesCompleteTable + """ c ON (f.id = c.fileid)
                WHERE m.run = sub.run_id and c.subscription = (SELECT id
                                                                  FROM """ + wmbsSubscriptionTable + """
                                                                  WHERE fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'AlcaSkimmable')
                                                                  )
                ) numAlcaskimmed,

            (SELECT  COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsSubFilesCompleteTable + """ c ON (f.id = c.fileid)
                WHERE  m.run = sub.run_id and c.subscription = (SELECT s.id
                                                                   FROM """ + wmbsSubscriptionTable + """ s INNER JOIN """ + wmbsSubsTypeTable + """ t ON (s.subtype = t.id)
                                                                   WHERE t.name = 'Exportable'
                                                                   )
                        AND f.id IN (SELECT f2.id
                                        FROM """ + wmbsFileDetailsTable + """ f2 INNER JOIN """ + wmbsDatasetAssocTable  + """ a2 ON (a2.file_id = f2.id)
                                                INNER JOIN """ + datasetTable  + """ dp2 ON (a2.dataset_path_id = dp2.id)
                                                INNER JOIN """ + dataTierTable + """ dt2 ON (dp2.data_tier = dt2.id)
                                                INNER JOIN """ + wmbsFileRunlumiMapTable + """ m2 ON (f2.id = m2.fileid)
                                                INNER JOIN """ + wmbsSubFilesCompleteTable + """ c2 ON (f2.id = c2.fileid)
                                        WHERE dt2.name = 'RAW' and  m2.run = sub.run_id and c2.subscription = (SELECT id
                                                                                                                FROM """ + wmbsSubscriptionTable + """
                                                                                                                WHERE fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Mergeable')
                                                                                                                )
                                        )
                ) numExportedRepackMerge,

            (SELECT  COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsSubFilesCompleteTable + """ c ON (f.id = c.fileid)
                WHERE  m.run = sub.run_id and c.subscription = (SELECT s.id
                                                                   FROM """ + wmbsSubscriptionTable + """ s INNER JOIN """ + wmbsSubsTypeTable + """ t ON (s.subtype = t.id)
                                                                   WHERE t.name = 'Exportable'
                                                                   )
			AND f.id IN (SELECT f2.id
					FROM """ + wmbsFileDetailsTable + """ f2
			                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m2 ON (f2.id = m2.fileid)
			                        INNER JOIN """ + wmbsSubFilesCompleteTable + """ c2 ON (f2.id = c2.fileid)
			                WHERE m2.run = sub.run_id and c2.subscription = (SELECT id
                        			                                            FROM """ + wmbsSubscriptionTable + """
                                                			                    WHERE fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Reconstructable')
                                                                                	    )
                                        )
                ) numExportedReco,

            (SELECT  COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsSubFilesCompleteTable + """ c ON (f.id = c.fileid)
                WHERE  m.run = sub.run_id and c.subscription = (SELECT s.id
                                                                   FROM """ + wmbsSubscriptionTable + """ s INNER JOIN """ + wmbsSubsTypeTable + """ t ON (s.subtype = t.id)
                                                                   WHERE t.name = 'Exportable'
                                                                   )
                        AND f.id IN (SELECT f2.id
                                        FROM """ + wmbsFileDetailsTable + """ f2
                                                INNER JOIN """ + wmbsFileRunlumiMapTable + """ m2 ON (f2.id = m2.fileid)
                                                INNER JOIN """ + wmbsSubFilesCompleteTable + """ c2 ON (f2.id = c2.fileid)
                                        WHERE m2.run = sub.run_id and c2.subscription = (SELECT id
                                                                                            FROM """ + wmbsSubscriptionTable + """
                                                                                            WHERE fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'AlcaSkimmable')
                                                                                            )
                                        )
                ) numExportedAlcaskim

            FROM(
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.hltkey hltkey,
                r.acq_era acq_era,
                rs.status status
                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order + """
            ) sub"""
        else:
            # slow query

	    #########
	    ## TODO: correct this query completly for the new schema. Is it still useful???
	    #########

            self.mainQuery = """
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.hltkey hltkey,
                r.acq_era acq_era,
                rs.status status,
                (SELECT COUNT(s.run_id) FROM """ + streamerTable + """ s WHERE s.run_id = r.run_id) numStreamers,
                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numRepacked,

                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numMerged,

                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numReconstructed,

                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numExportedRepacked,

                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numExportedReco


                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order

#        globals.logger.log(self.query)
